#include "CivHydrSimulation.h"
#include "CivPgConn.h"
#include "CivNewInp.h"
#include "CivBuilder.h"
#include "CivInpDirector.h"
#include "CivSysLog.h"
#include "epanet2.h"
#include "../pandaDbManager/CivCommonUtils.h"


CivHydrSimulation::CivHydrSimulation()
   :mDbConn(nullptr),
	mInpFile("test.inp"),
	mRptFile("test.rpt"),
	mBinFile("test.bin")
{

}

CivHydrSimulation::CivHydrSimulation(CivDbConn* dbConn)
	:mDbConn(dbConn),
	mInpFile("test.inp"),
	mRptFile("test.rpt"), 
	mBinFile("test.bin")
{

}

CivHydrSimulation::~CivHydrSimulation()
{

}

char* CivHydrSimulation::getInpFile()
{ 
	return const_cast<char*>(mInpFile.c_str()); 
}

void CivHydrSimulation::setInpFile(const std::string& inpFile)
{ 
	mInpFile = inpFile; 
}

char* CivHydrSimulation::getRptFile()
{ 
	return const_cast<char*>(mRptFile.c_str()); 
}

void CivHydrSimulation::setRptFile(const std::string& rptFile)
{ 
	mRptFile = rptFile; 
}

char* CivHydrSimulation::getBinFile()
{
	return const_cast<char*>(mBinFile.c_str());
}

void CivHydrSimulation::setBinFile(const std::string& binFile)
{ 
	mBinFile = binFile; 
}

void CivHydrSimulation::registDb(CivDbConn* dbConn)
{
	if (dbConn == nullptr)
		return;

	if (dbConn != nullptr)
	{
		dbConn->close();
		delete mDbConn;
		mDbConn = nullptr;
	}
	mDbConn = dbConn;
}

bool CivHydrSimulation::convertInp()
{
	CivBuilder* builder = new CivInpBuilder();
	builder->registDb(mDbConn);

	CivInpDirector director;
	director.setBuilder(builder);
	director.create();

	CivNewInp* newInp = builder->getNewInp();
	char* inpFileName = getInpFile();

	if (inpFileName 
		&& (*inpFileName !='\0') 
		&& !newInp->writeToFile(inpFileName))
	{
		delete builder;
		delete newInp;
		return false;
	}

	delete builder;
	delete newInp;
	return true;
}

int CivHydrSimulation::hdyrSimulation()
{
	CivSysLog::getInstance()->error("ʼˮ", "CivHydrSimulation", __FUNCTION__);
	if (!convertInp())
	{
		CivSysLog::getInstance()->error("дinpļʧ", "CivHydrSimulation", __FUNCTION__);
		return false;
	}
	CivSysLog::getInstance()->info("дinpɹ", "CivHydrSimulation", __FUNCTION__);

    short hydCode = 1;
    long  t(0), tstep(0);
    int   iTime(0);

    int  errcode = 0;

    char* fileName = getInpFile();
    char* rptFile = getRptFile();
    char* binOutFile = getBinFile();

    errcode = ENopen(fileName, rptFile, binOutFile);
    if (errcode > 0)
    {
        ENclose();
        CivSysLog::getInstance()->error("ENopen inp ʧ", "CivHydrCalc", __FUNCTION__);
        return false;
    }
    CivSysLog::getInstance()->info("ENopen inp ɹ", "CivHydrCalc", __FUNCTION__);
    if (ENopenH() > 0)
    {
        ENclose();
        CivSysLog::getInstance()->error("ENopenH inp ʧ", "CivHydrCalc", __FUNCTION__);
        return false;
    }

    if (ENinitH(0) > 0)
    {
        ENclose();
        CivSysLog::getInstance()->error("ENinitH inp ʧ", "CivHydrCalc", __FUNCTION__);
        return false;
    }
    CivSysLog::getInstance()->info("ENinitH inp ɹ", "CivHydrCalc", __FUNCTION__);

    float dHStep;
    ENGetVal(4, &dHStep);

    bool isFirst = true;
    // 
    do
    {
        int nLinkCount;
        ENgetcount(EN_LINKCOUNT, &nLinkCount);
        /*for (int i = 1; i <= nLinkCount; i++)
        {
            int linkType;
            char id[128] = "";
            ENgetlinktype(i, &linkType);
            ENgetlinkid(i, id);
        }*/

        if (ENrunH(&t) > 100)	// errcode > 100 Ǵ
        {
            ENclose();
            return false;
        }
        ENnextH(&tstep);
        if (isFirst)
        {
            mDbConn->deleteByField(NODERESULTTABLE, "", CurrentDate);
            mDbConn->deleteByField(PIPERESULTTABLE, "", CurrentDate);
            isFirst = false;
        }
        // ȡǰڵ͹ܶ
        getNodeResult(iTime);
        getLinkResult(iTime);
        iTime++;
    } while (tstep > 0);

    // 
    if (!mResultCache.updateToDb(mDbConn))
    {
        CivSysLog::getInstance()->error("ˮ洢ʧ", "CivHydrCalc", __FUNCTION__);
        return false;
    }
    // ¹ݿ
    (void)mDbConn->updateNodeByInterval(CurrentDate, "0");
    (void)mDbConn->updateLinkByInterval(CurrentDate, "0");

    ENcloseH();
    ENclose();
    return true;
}

int CivHydrSimulation::qualitySimulation()
{
    CivSysLog::getInstance()->error("ʼˮʷ", "CivHydrCalc", __FUNCTION__);
    if (!convertInp())
    {
        CivSysLog::getInstance()->error("дinpļʧ", "CivHydrCalc", __FUNCTION__);
        return false;
    }
    CivSysLog::getInstance()->info("дinpɹ", "CivHydrCalc", __FUNCTION__);


    // ˮҪȽˮ,ûֱӲENSolveH Ϊ˿Ǽѹѹ
    long  t(0), tstep(0);
    int   iTime(0);

    int  errcode = 0;

    char* fileName = getBinFile();
    char* rptFile = getRptFile();
    char* binOutFile = getBinFile();

    errcode = ENopen(fileName, rptFile, binOutFile);
    if (errcode > 0)
    {
        ENclose();
        CivSysLog::getInstance()->error("ENopen inp ʧ", "CivHydrCalc", __FUNCTION__);
        return -1;
    }
    CivSysLog::getInstance()->info("ENopen inp ɹ", "CivHydrCalc", __FUNCTION__);
    if (ENopenH() > 0)
    {
        ENclose();
        CivSysLog::getInstance()->error("ENopenH inp ʧ", "CivHydrCalc", __FUNCTION__);
        return -1;
    }

    if (ENinitH(EN_SAVE) > 0)
    {
        ENclose();
        CivSysLog::getInstance()->error("ENinitH inp ʧ", "CivHydrCalc", __FUNCTION__);
        return -1;
    }
    CivSysLog::getInstance()->info("ENinitH inp ɹ", "CivHydrCalc", __FUNCTION__);
    bool isFirst = true;
    do
    {
        int nLinkCount;
        ENgetcount(EN_LINKCOUNT, &nLinkCount);
        /*for (int i = 1; i <= nLinkCount; i++)
        {
            int linkType;
            char id[128] = "";
            ENgetlinktype(i, &linkType);
            ENgetlinkid(i, id);
        }*/

        errcode = ENrunH(&t);	// errcode > 100 Ǵ
        if (errcode > 100)
        {
            ENopenQ();
            ENclose();
            return -1;
        }
        if (isFirst)
        {
            mDbConn->deleteByField(NODERESULTTABLE, "", CurrentDate);
            mDbConn->deleteByField(PIPERESULTTABLE, "", CurrentDate);
            isFirst = false;
        }
        // ȡǰڵ͹ܶ
        getNodeResult(iTime);
        getLinkResult(iTime);
        ENnextH(&tstep);

        iTime++;

    } while (tstep > 0);

    ENcloseH();

    // ˮʷ
    ENopenQ();
    ENinitQ(1);

    t = 0;
    tstep = 0;
    iTime = 0;
    do
    {
        errcode = ENrunQ(&t);	// errcode > 100 Ǵ
        if (errcode > 100)
        {
            ENcloseQ();
            ENclose();
            return -1;
        }

        // ȡˮģ
        getNodeQuality(iTime);
        getLinkQuality(iTime);

        ENstepQ(&tstep);

        iTime++;

    } while (tstep > 0);


    ENcloseQ();
    ENclose();

    // 
    if (!mResultCache.updateToDb(mDbConn))
    {
        CivSysLog::getInstance()->error("ˮʷ洢ʧ", "CivHydrCalc", __FUNCTION__);
        return -1;
    }
    CivSysLog::getInstance()->info("ˮʷ洢ɹ", "CivHydrCalc", __FUNCTION__);

    // ¹ݿ
    (void)mDbConn->updateNodeByInterval(CurrentDate, "0");
    (void)mDbConn->updateLinkByInterval(CurrentDate, "0");

    CivSysLog::getInstance()->info("ˮʷ", "CivHydrCalc", __FUNCTION__);
    return 1;
}

int CivHydrSimulation::trackingSimulation(char* snNode)
{
    return 1;
}

void CivHydrSimulation::getNodeResult(short time)
{
    int nNodeCount;
    ENgetcount(EN_NODECOUNT, &nNodeCount);

    for (int i = 1; i <= nNodeCount; i++)
    {
        NodeResultItem nodeItem;
        // 
        ENgetnodeid(i, nodeItem.szNo);
        // 
        ENgetnodevalue(i, EN_DEMAND, &nodeItem.dFlow);
        // ˮͷ
        ENgetnodevalue(i, EN_HEAD, &nodeItem.dHead);
        // ѹ
        ENgetnodevalue(i, EN_PRESSURE, &nodeItem.dPressure);
        // 
        ENgetnodevalue(i, EN_ELEVATION, &nodeItem.dElevation);
        // ʵˮ
        ENgetnodevalue(i, EN_DEMAND, &nodeItem.dDemand);
        // ˮ
        ENgetnodevalue(i, EN_TANKVOLUME, &nodeItem.dTankVolume);
        // ˮ
        ENgetnodevalue(i, EN_MAXVOLUME, &nodeItem.dTankMaxVolume);
        // Һλ߶
        ENgetnodevalue(i, EN_TANKLEVEL, &nodeItem.dTankLevel);

        nodeItem.dInterval = time;

        // ŵ 
        mResultCache.addNodeItems(nodeItem);
    }
}

void CivHydrSimulation::getLinkResult(short time)
{
    int  nLinkCount;
    ENgetcount(EN_LINKCOUNT, &nLinkCount);

    for (int i = 1; i <= nLinkCount; i++)
    {
        LinkResultItem linkItem;

        int* typeCode = (int*)malloc(sizeof(int));

        ENgetlinktype(i, typeCode);
        if (*typeCode != EN_PIPE)
        {
            free(typeCode);
            continue;
        }
        free(typeCode);

        // 
        ENgetlinkid(i, linkItem.szNo);
        //  
        ENgetlinkvalue(i, EN_FLOW, &linkItem.dFlow);
        // 
        linkItem.nFlowDirect = (linkItem.dFlow >= 0) ? 0 : 1;
        // 
        ENgetlinkvalue(i, EN_VELOCITY, &linkItem.dVelocity);
        // ˮͷʧ
        ENgetlinkvalue(i, EN_HEADLOSS, &linkItem.dHeadloss);
        // λˮͷʧ
        float dPipeLen = 0.0f;;
        ENgetlinkvalue(i, EN_LENGTH, &dPipeLen);
        linkItem.dUnitHeadloss = (linkItem.dHeadloss / dPipeLen) * 1000;
        // ȡϵͱ
        int node1 = 0;
        int node2 = 0;
        ENgetlinknodes(i, &node1, &node2);
        // ϵˮͷ
        ENgetnodevalue(node1, EN_HEAD, &linkItem.dFromNodHeadloss);
        // ˮͷ
        ENgetnodevalue(node2, EN_HEAD, &linkItem.dToNodHeadloss);

        // ֲˮͷʧ
        float fCoeff;
        ENgetlinkvalue(i, EN_MINORLOSS, &fCoeff);
        double dMinorloss = fCoeff * linkItem.dHeadloss;
        linkItem.dLocalHeadloss = dMinorloss;// ֲˮͷʧ 

        // Ħˮͷʧ
        linkItem.dFrictionHeadloss = linkItem.dHeadloss - linkItem.dLocalHeadloss;// ֲˮͷʧ 

        // ״̬
        float fState = 0;
        ENgetlinkvalue(i, EN_STATUS, &fState);
        if ((int)fState == 0)
            strcpy_s(linkItem.szStatus, "ر");
        else
            strcpy_s(linkItem.szStatus, "");

        linkItem.dInterval = time;
        // ŵ 
        mResultCache.addLinkItems(linkItem);
    }
}

void CivHydrSimulation::getNodeQuality(short time)
{
    int nNodeCount;
    ENgetcount(EN_NODECOUNT, &nNodeCount);


    for (int i = 1; i <= nNodeCount; i++)
    {
        char No[35];
        // 
        ENgetnodeid(i, No);
        float* quality = (float*)malloc(sizeof(float));
        // ˮ
        ENgetnodevalue(i, EN_QUALITY, quality);

        mResultCache.addNodeQuality(*quality, time, No);
        free(quality);
    }
}

void CivHydrSimulation::getLinkQuality(short time)
{
    int  nLinkCount;
    ENgetcount(EN_LINKCOUNT, &nLinkCount);
    int* typeCode = (int*)malloc(sizeof(int));
    float* quality = (float*)malloc(sizeof(float));
    for (int i = 1; i <= nLinkCount; i++)
    {

        ENgetlinktype(i, typeCode);
        if (*typeCode != EN_PIPE)
        {
            continue;
        }
        char lnkNo[35];
        // 
        ENgetlinkid(i, lnkNo);

        // ˮ
        ENgetlinkvalue(i, EN_LINKQUAL, quality);
        mResultCache.addLinkQuality(*quality, time, lnkNo);
    }
    free(typeCode);
    free(quality);
}